def simulate_clean_experiment():
    """1. przenieść src\\qrem\\usecases\\mitigation\\create_mitigation_data_KKM.py funkcja: simulate_experiment do  
    src\\qrem\\common\\providers\\simulation.py->simulate_clean_experiment()
    """
    pass


def simulate_noisy_experiment():
    """2.stworzyć funkcję simulate_noisy_experiment() w \\src\\qrem\\common\\providers\\simulation.py, która na podstawie wyników simulate_clean_experiment i parametru noise_model="CN" i innych zwróci noisy_results, czyli np. w przypadku CN wywoa cunkcję:
	src\\qrem\\cn\\simulation.py::simulate_noise_results_dictionary()
    """

    pass