"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
#TODO FBM: think whether do something here. For now, check noise_characterization.data_analysis.InitialNoiseAnalyzer
#ORGANIZE - this file seams obsolete (MO)