"""4. napisać funkcję run_simulation_experiment.py, która (1) stworzy przykladowy model(e), (2) zwróci nam jakąś przykladową symulację."""

from qrem.common.printer import qprint, errprint, warprint



if __name__ == "__main__":
    warprint("=============================")
    warprint("START: Executing SIMULATION Experiment pipeline")
    warprint("=============================")